import pytest
from bddrest import status, response, given

from nanohttp import Controller, action, html, json, text, xml, binary
from nanohttp.tests.helpers import Given, when


def test_action_decorator_verbs():
    class Root(Controller):
        @action(verbs=['get', 'view'])
        def index(self, a, b):
            yield f'{a}, {b}'

    with Given(Root(), '/1/2'):
        assert status == 200
        assert response.text == '1, 2'

        when(verb='put')
        assert status  == 405


def test_action_decorator_content_type():
    class Root(Controller):
        @action(encoding='utf32', content_type='text/plain')
        def index(self, a, b):
            yield f'{a}, {b}'

    with Given(Root(), '/1/2'):
        assert status == 200
        assert response.encoding == 'utf32'
        assert response.content_type == 'text/plain'
        assert response.text == '1, 2'


def test_action_decorator_prevent_form():
    class Root(Controller):
        @action(prevent_form=True)
        def index(self):
            yield 'default'

        @action(prevent_form='777 Form Not Allowedi Here')
        def custom_status(self):
            yield 'custom'

    with Given(Root(), verb='POST'):
        assert status == 200

        when(form=dict(a=1))
        assert status == 400

        when('/custom_status', form=dict(a=1))
        assert status == 777

        when('/custom_status')
        assert status == 200


def test_action_decorator_prevent_empty_form():
    class Root(Controller):
        @action(prevent_empty_form=True)
        def index(self):
            yield 'default'

        @action(prevent_empty_form='777 Form Is Required Here')
        def custom_status(self):
            yield 'custom'

    with Given(Root(), verb='POST', form=dict(a=1)):
        assert status == 200

        when(form={})
        assert status == 400

        when(form=None)
        assert status == 400

        when('/custom_status', form={})
        assert status == 777

        when('/custom_status', form=None)
        assert status == 777

        when('/custom_status', form=dict(a=1))
        assert status == 200


def test_action_decorator_form_whitelist():
    class Root(Controller):
        @action(form_whitelist=['a', 'b'])
        def index(self):
            yield 'default'

        @action(form_whitelist=(['a', 'b'], '888 Only a & b accepted'))
        def custom_status(self):
            yield 'custom'

    with Given(Root(), verb='POST', form=dict(a=1)):
        assert status == 200

        when(form=given + dict(b=2))
        assert status == 200

        when(form=given + dict(c=2))
        assert status == 400

        when('/custom_status', form=dict(c=2))
        assert status == 888

        when('/custom_status', form=dict(a=2))
        assert status == 200


def test_html_decorator():
    class Root(Controller):
        @html
        def index(self):
            yield '<html></html>'

    with Given(Root()):
        assert status == 200
        assert response.text == '<html></html>'
        assert response.content_type == 'text/html'
        assert response.encoding == 'utf-8'


def test_json_decorator():
    class Root(Controller):
        @json
        def index(self, a, b):
            return dict(a=a, b=b)

        @json
        def custom(self):
            class A:
                def to_dict(self):
                    return dict(c=1)
            return A()

        @json
        def badobject(self):
            class A:
                pass
            return A()

    with Given(Root(), '/1/2'):
        assert status == 200
        assert response.json == dict(a='1', b='2')
        assert response.content_type == 'application/json'
        assert response.encoding == 'utf-8'

        when('/custom')
        assert status == 200
        assert response.json == dict(c=1)

        with pytest.raises(ValueError):
            when('/badobject')


def test_text_decorator():
    class Root(Controller):
        @text
        def index(self):
            yield 'abc'

    with Given(Root()):
        assert status == 200
        assert response.text == 'abc'
        assert response.content_type == 'text/plain'
        assert response.encoding == 'utf-8'


def test_xml_decorator():
    class Root(Controller):
        @xml
        def index(self):
            yield '<xml></xml>'

    with Given(Root()):
        assert status == 200
        assert response.text == '<xml></xml>'
        assert response.content_type == 'application/xml'
        assert response.encoding == 'utf-8'


def test_binary_decorator():
    class Root(Controller):
        @binary
        def index(self):
            yield b'abc'

    with Given(Root()):
        assert status == 200
        assert response.body == b'abc'
        assert response.content_type == 'application/octet'
        assert response.encoding is None




import functools

import bddrest

from nanohttp import Application, configure, settings


def Given(controller, *args, configuration=None, **kwargs):
    if configuration:
        configure(configuration, force=True)
    else:
        configure(force=True)

    if isinstance(controller, Application):
        application = controller
    else:
        application = Application(controller)
    return bddrest.Given(application, None, *args, **kwargs)


when = functools.partial(bddrest.when, None)





import sys
import types
import logging
import traceback
from collections import Iterable

from nanohttp.contexts import Context, context
from nanohttp.exceptions import HTTPStatus
from nanohttp.configuration import settings
from nanohttp.constants import NO_CONTENT_STATUSES


logger = logging.getLogger('nanohttp')


class Application:
    """Application main handler
    """

    #: Application logger based on python builtin logging module
    __logger__ = logger

    #: The root controller
    __root__ = None

    def __init__(self, root=None):
        """Initialize application and calling ``app_init`` hook.

        .. note:: ``__root__`` attribute will set by ``root`` parameter.

        :param root: The root controller
        """
        self.__root__ = root
        self._hook('app_init')

    def _hook(self, name, *args, **kwargs):
        """Call the hook

        :param name: Hook name
        :param args: Pass to the hook positional arguments
        :param kwargs: Pass to the hook keyword arguments
        """
        if hasattr(self, name):
            return getattr(self, name)(*args, **kwargs)

    def _handle_exception(self, ex, start_response):

        response_headers = [("content-type", "text/plain")],
        if isinstance(ex, HTTPStatus):
            exc_info=None
            status = ex.status
            response_body = ex.render()
            if ex.headers:
                response_headers = ex.headers
        else:
            self.__logger__.exception(
                'Internal Server Error',
                exc_info=True
            )
            exc_info=sys.exc_info()
            status = '500 Internal Server Error'
            if settings.debug:
                response_body = traceback.format_exc()
            else:
                response_body = status

        start_response(
            status,
            response_headers,
            exc_info=exc_info
        )
        self._hook('end_response')
        context.__exit__(*sys.exc_info())

        # Sometimes don't need to transfer any body, for example the 304 case.
        if status[:3] in NO_CONTENT_STATUSES:
            return []

        return [response_body.encode()]

    def __call__(self, environ, start_response):
        """Method that
        `WSGI <https://www.python.org/dev/peps/pep-0333/#id15>`_ server calls
        """
        # Entering the context
        context_ = Context(environ, self)
        context_.__enter__()

        # Preparing some variables
        status = '200 OK'
        buffer = None
        response_iterable = None

        try:
            self._hook('begin_request')

            # Removing the trailing slash in-place, if exists
            context_.path = context_.path.rstrip('/')

            # Removing the heading slash, and query string anyway
            path = context_.path[1:].split('?')[0]

            # Splitting the path by slash(es) if any
            remaining_paths = path.split('/') if path else []

            # Calling the controller, actually this will be serve our request
            response_body = self.__root__(*remaining_paths)

            if response_body:
                # The goal is to yield an iterable, to encode and iter over it
                # at the end of this method.

                if isinstance(response_body, (str, bytes)):
                    # Mocking the body inside an iterable to prevent
                    # the iteration over the str character by character
                    # For more info check the pull-request
                    # #34, https://github.com/Carrene/nanohttp/pull/34
                    response_iterable = (response_body, )

                elif isinstance(response_body, types.GeneratorType):
                    # Generators are iterable !
                    response_iterable = response_body

                    # Trying to get at least one element from the generator,
                    # to force the method call till the second
                    # `yield` statement
                    buffer = next(response_iterable)

                elif isinstance(response_body, Iterable):
                    # Creating an iterator from iterable!
                    response_iterable = iter(response_body)

                else:
                    raise ValueError(
                        'Controller\'s action/handler response must be '
                        'generator and or iterable'
                    )

        except Exception as ex:
            return self._handle_exception(ex, start_response)

        self._hook('begin_response')

        # Setting cookies in response headers, if any
        cookie = context_.cookies.output()
        if cookie:
            for line in cookie.split('\r\n'):
                context_.response_headers.add_header(*line.split(': ', 1))

        start_response(
            status,
            context_.response_headers.items(),
        )

        # It seems we have to transfer a body, so this function should yield
        # a generator of the body chunks.
        def _response():
            try:
                if buffer is not None:
                    yield context_.encode_response(buffer)

                if response_iterable:
                    # noinspection PyTypeChecker
                    for chunk in response_iterable:
                        yield context_.encode_response(chunk)
                else:
                    yield b''
            except Exception as ex_:
                self.__logger__.exception(
                    'Exception while serving the response.'
                )
                if settings.debug:
                    yield str(ex_).encode()
                raise ex_

            finally:
                self._hook('end_response')
                context.__exit__(*sys.exc_info())

        return _response()

    def shutdown(self):  # pragma: nocover
        pass




import functools
from inspect import signature, Parameter
from typing import Union

import ujson

from .contexts import context
from .constants import UNLIMITED


def action(*args, verbs='any', encoding='utf-8', content_type=None,
           inner_decorator=None, prevent_empty_form=None, prevent_form=None,
           form_whitelist=None, **kwargs):
    """
    Base action decorator

    Marks the function as a nanohttp handler/action.

    :param verbs: Allowed HTTP methods as list or tuple of strings.
    :param encoding: Content encoding
    :param content_type: Response Content Type
    :param inner_decorator: Inner decorator, to put it between this decorator
                            and the handler.
    :param prevent_empty_form: Boolean or str, indicates to prevent empty HTTP
                               form. if str given, a
                               :class:`.HTTPStatus(<str>)` will be raised.
                               otherwise :class:`.HTTPBadRequest`.
    :param prevent_form: Boolean or str, indicates to prevent any HTTP form. if
                         str given, a :class:`.HTTPStatus(<str>)` will be
                         raised, otherwise :class:`.HTTPBadRequest`.
    :param form_whitelist: A list of allowed form fields. or a
                           tuple(list, httpstatus)
    """
    def decorator(func):
        nonlocal verbs

        if inner_decorator is not None:
            func = inner_decorator(func, *args, **kwargs)

        # Examining the signature,
        # and counting the optional and positional arguments.
        positional_arguments, optional_arguments, keywordonly_arguments = \
            [], [], []
        action_signature = signature(func)
        positional_unlimited = False
        optional_unlimited = False

        for name, parameter in action_signature.parameters.items():
            if name == 'self':
                continue

            if parameter.kind == Parameter.KEYWORD_ONLY:
                keywordonly_arguments.append(
                    (parameter.name, parameter.default)
                )
            elif parameter.kind == Parameter.VAR_POSITIONAL:
                positional_unlimited = True

            elif parameter.kind == Parameter.VAR_KEYWORD:
                optional_unlimited = True

            elif parameter.default is Parameter.empty:
                positional_arguments.append(parameter.name)
            else:
                optional_arguments.append((parameter.name, parameter.default))

        func.__nanohttp__ = dict(
            verbs=[verbs] if isinstance(verbs, str) else verbs,
            encoding=encoding,
            content_type=content_type,
            minimum_allowed_arguments=len(positional_arguments),
            maximum_allowed_arguments= \
                UNLIMITED if positional_unlimited else \
                len(positional_arguments) + len(optional_arguments),
            keywordonly_arguments=keywordonly_arguments,
            prevent_empty_form=prevent_empty_form,
            prevent_form=prevent_form,
            form_whitelist=form_whitelist,
            default_action='index'
        )

        return func

    if args and callable(args[0]):
        f = args[0]
        args = tuple()
        return decorator(f)

    return decorator


def chunked(trailer_field=None, trailer_value=None):
    """Enables chunked encoding on an action
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal trailer_field, trailer_value
            context.response_headers.add_header('Transfer-Encoding', 'chunked')
            if trailer_field:
                context.response_headers.add_header('Trailer', trailer_field)
            result = func(*args, **kwargs)
            try:
                while True:
                    chunk = next(result)
                    yield f'{len(chunk)}\r\n{chunk}\r\n'

            except StopIteration:
                yield '0\r\n'
                if trailer_field and trailer_value:
                    yield f'{trailer_field}: {trailer_value}\r\n'
                yield '\r\n'

            except Exception as ex:
                exstr = str(ex)
                yield f'{len(exstr)}\r\n{exstr}'
                yield '0\r\n\r\n'

        return wrapper

    if callable(trailer_field):
        func = trailer_field
        trailer_field = None
        return decorator(func)

    return decorator


def jsonify(func):

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if hasattr(result, 'to_dict'):
            result = result.to_dict()
        elif not isinstance(result, (list, dict, int, str)):
            raise ValueError('Cannot encode to json: %s' % type(result))

        return ujson.dumps(result, indent=4)

    return wrapper


#: HTML action decorator
html = functools.partial(action, content_type='text/html')

#: Plain Text action decorator
text = functools.partial(action, content_type='text/plain')

#: JSON action decorator
#:
#: accepts list, dict, int, str or objects have ``to_dict`` method.
json = functools.partial(
    action,
    content_type='application/json',
    inner_decorator=jsonify
)

#: XML action decorator
xml = functools.partial(action, content_type='application/xml')

#: Binary-data action decorator
binary = functools.partial(
    action,
    content_type='application/octet-stream',
    encoding=None
)




import re
from decimal import Decimal

import pytest
from nanohttp import RequestValidator, HTTPBadRequest, HTTPStatus, \
    Controller, validate, action, context
from bddrest import status, response

from nanohttp.tests.helpers import Given, when


def test_validation_decorator():
    class Root(Controller):

        @validate(
            query1=dict(
                required=True,
                type_=int,
                query=True
            ),
            field1=dict(
                required=True,
                type_=float
            ),
            field2=dict(
                required=True,
                type_=int
            ),
            field3=dict(
                required=True,
                type_=lambda v: v.encode(),
                min_length=1
            )
        )
        @action
        def index(self):
            query1 = context.query['query1']
            field1 = context.form['field1']
            field2 = context.form['field2']
            field3 = context.form['field3']
            yield \
                f'{type(query1).__name__}: {query1}, '\
                f'{type(field1).__name__}: {field1}, ' \
                f'{type(field2).__name__}: {field2}, ' \
                f'{type(field3).__name__}: {field3}'

    with Given(
        Root(),
        verb='POST',
        query=dict(query1=1),
        form=dict(field1='2.3', field2='2', field3='ab'),
    ):
        assert status == 200
        assert response.text == 'int: 1, float: 2.3, int: 2, bytes: b\'ab\''


def test_validation_required():

    validator = RequestValidator(
        fields=dict(
            a=dict(required=True)
        )
    )

    # Trying to pass with a
    assert dict(a='value1') == validator(dict(a='value1'))[0]

    # Trying to pass without a
    with pytest.raises(HTTPBadRequest):
        validator(dict(another_param='value1'))

    # Define required validation with custom exception
    validator = RequestValidator(
        fields=dict(
            a=dict(required='600 Custom exception')
        )
    )

    # Trying to pass with another param without a
    with pytest.raises(HTTPStatus('600 Custom exception').__class__):
        validator(dict(another_param='value1'))

    # Trying to pass with a
    assert dict(a='value1') == validator(dict(a='value1'))[0]


def test_validation_max():

    validator = RequestValidator(
        fields=dict(
            a=dict(maximum=10)
        )
    )
    assert dict(a=9) == validator(dict(a=9))[0]
    assert dict(a=None) == validator(dict(a=None))[0]

    with pytest.raises(HTTPBadRequest):
         validator(dict(a='9'))

    # More than Expectation
    with pytest.raises(HTTPBadRequest):
        validator(dict(a=11))

    # Not int input
    with pytest.raises(HTTPBadRequest):
        validator(dict(a='a'))


def test_validation_min():

    validator = RequestValidator(
        fields=dict(
            a=dict(minimum=10)
        )
    )

    assert dict(a=11) == validator(dict(a=11))[0]
    with pytest.raises(HTTPBadRequest):
         validator(dict(a='11'))

    # Less than Expectation
    with pytest.raises(HTTPBadRequest):
        validator(dict(a=9))


def test_validation_min_length():

    validator = RequestValidator(
        fields=dict(
            a=dict(min_length=3)
        )
    )
    assert dict(a='abc') == validator(dict(a='abc'))[0]

    # Shorter than Expectation
    with pytest.raises(HTTPBadRequest):
        validator(dict(a='ab'))


def test_validation_max_length():
    validator = RequestValidator(
        fields=dict(
            a=dict(max_length=4)
        )
    )
    assert dict(a='abcd') == validator(dict(a='abcd'))[0]

    # Longer than Expectation
    with pytest.raises(HTTPBadRequest):
        validator(dict(a='abbcde'))


def test_validation_pattern():
    validator = RequestValidator(
        fields=dict(
            a=dict(pattern=r'^\D{5}$')
        )
    )

    assert dict(a='abcdJ') == validator(dict(a='abcdJ'))[0]

    # Param1 not matching
    with pytest.raises(HTTPBadRequest):
        validator(dict(a='asc'))


def test_validation_pattern_compiled():
    validator = RequestValidator(
        fields=dict(
            a=dict(pattern=re.compile(r'^\d{5}$'))
        )
    )

    assert dict(a='01234') == validator(dict(a='01234'))[0]

    # Param1 not matching
    with pytest.raises(HTTPBadRequest):
        validator(dict(a='123456'))


def test_validation_type():
    validator = RequestValidator(
        fields=dict(
            a=dict(type_=int),
            b=dict(type_=str),
            c=dict(type_=Decimal)
        )
    )

    assert dict(a=123, b='123', c=1.23) == \
        validator(dict(a='123', b=123, c=1.23))[0]

    # Param1 bad value(Cannot be converted to int)
    with pytest.raises(HTTPBadRequest):
        validator(dict(
            a='NotInteger',
        ))

    # Param3 bad value(Cannot be converted to decimal)
    with pytest.raises(HTTPBadRequest):
        validator(dict(c='NotDecimal'))


def test_validation_query_string():
    validator = RequestValidator(
        fields=dict(
            a=dict(query=True),
        )
    )

    # Accept query_string
    assert dict(a='value') == validator(query=dict(a='value'))[1]


def test_validation_custom_status():
    validator = RequestValidator(
        fields=dict(
            a=dict(
                type_=(int, '999 Type error'),
                minimum=(3, '666'),
                maximum=(4, '667 greater than maximum'),
            )
        )
    )

    with pytest.raises(HTTPStatus) as ctx:
        validator(dict(a='NotInteger'))
    assert str(ctx.value) == '999 Type error'

    with pytest.raises(HTTPStatus) as ctx:
        validator(dict(a=2))
    assert str(ctx.value) == '666 Bad request'

    with pytest.raises(HTTPStatus) as ctx:
        validator(dict(a=5))
    assert str(ctx.value) == '667 greater than maximum'


def test_callable_validator():
    def f(age, container, field):
        age = int(age)
        if age < 0:
            raise ValueError('Age is not in valid range')
        return age

    validator = RequestValidator(
        fields=dict(
            a=f
        )
    )

    assert dict(a=12).items() == validator(dict(a=12))[0].items()

    with pytest.raises(ValueError):
        validator(dict(a=-1))


def test_not_none_validator():
    validator = RequestValidator(fields=dict(a=dict(not_none=True)))
    assert dict(a=1, b=2) == validator(dict(a=1, b=2))[0]
    assert dict(b=2) == validator(dict(b=2))[0]

    with pytest.raises(HTTPBadRequest):
        validator(dict(a=None))

    validator = RequestValidator(
        fields=dict(a=dict(not_none='666 a is null'))
    )
    with pytest.raises(HTTPStatus) as ctx:
        validator(dict(a=None))
    assert '666 a is null' == str(ctx.value)

    with pytest.raises(TypeError):
        RequestValidator(fields=dict(a=dict(not_none=23)))


def test_readonly_validator():
    validator = RequestValidator(fields=dict(a=dict(readonly=True)))
    with pytest.raises(HTTPBadRequest):
        validator(dict(a=None))

    validator = RequestValidator(
        fields=dict(a=dict(not_none='666 a is readonly'))
    )
    with pytest.raises(HTTPStatus) as ctx:
        validator(dict(a=None))
    assert '666 a is readonly' == str(ctx.value)




from .exceptions import HTTPStatus, HTTPBadRequest, HTTPUnauthorized, \
    HTTPForbidden, HTTPNotFound, HTTPMethodNotAllowed, HTTPConflict, HTTPGone,\
    HTTPRedirect, HTTPMovedPermanently, HTTPFound, HTTPInternalServerError, \
    HTTPNotModified, HTTPBadGatewayError, HTTPCreated, HTTPAccepted,\
    HTTPNonAuthoritativeInformation, HTTPNoContent, HTTPResetContent,\
    HTTPPartialContent, HTTPKnownStatus
from .controllers import Controller, RestController, Static, \
    RegexRouteController
from .decorators import action, html, json, xml, binary, text, chunked
from .helpers import quickstart, LazyAttribute
from .cli import main
from .contexts import context, ContextIsNotInitializedError
from .configuration import settings, configure
from .application import Application
from .validation import validate, RequestValidator

__version__ = '1.9.7'



from bddrest import status, response

from nanohttp import Static
from nanohttp.tests.helpers import Given, when


def test_static_controller(make_temp_directory):
    static = Static(
        make_temp_directory(a=dict(a1='A1', a2='A2'), b='B'),
        default_document='b'
    )
    with Given(static):
        assert status == 200
        assert response.text == 'B'

        when('/a/a1')
        assert status == 200
        assert response.text == 'A1'

        when('/a/a2')
        assert status == 200
        assert response.text == 'A2'

        when('/a/a3')
        assert status == 404

        when('/../')
        assert status == 403

        when('/a/../a/a2')
        assert status == 200

        # When default document is not exists
        static.default_document = 'BadFile'
        when('/')
        assert status == 404

        # When default document is not given
        static.default_document = None
        when('/')
        assert status == 404




from bddrest import status, response

from nanohttp import Controller, action
from nanohttp.tests.helpers import Given, when


def test_fixed_positional_url_parameter():
    class Root(Controller):
        @action
        def index(self, a, b):
            yield f'{a}, {b}'

    with Given(Root(), '/1/2'):
        assert status == 200
        assert response.text == '1, 2'

        # Extra trailing slash
        when('/1/2/')
        assert status == 200
        assert response.text == '1, 2'

        # Insufficient URL parameters
        when('/1')
        assert status == 404

        # Extra URL parameters
        when('/1/2/3')
        assert status  == 404


def test_vary_positional_url_parameters():
    class Root(Controller):
        @action
        def index(self, *args):
            yield ', '.join(args)

    with Given(Root()):
        assert status == 200
        assert response.text == ''

        when('/a')
        assert status == 200
        assert response.text == 'a'

        when('/a/b')
        assert status == 200
        assert response.text == 'a, b'


def test_fixed_optional_arguments_url_parameters():
    class Root(Controller):
        @action
        def index(self, a=None, b=1):
            yield f'{a}, {b}'

    with Given(Root()):
        assert status == 200
        assert response.text == 'None, 1'

        when('/a')
        assert status == 200
        assert response.text == 'a, 1'

        when('/a/b')
        assert status == 200
        assert response.text == 'a, b'

        when('/a/b/c')
        assert status == 404


def test_vary_optional_arguments_url_parameters():
    class Root(Controller):
        @action
        def index(self, **kw):
            yield ', '.join(f'{k}={v}' for k, v in kw.items())

    with Given(Root()):
        assert status == 200
        assert response.text == ''

        when('/a')
        assert status == 404


def test_both_vary_optional_and_positional_arguments_url_parameters():
    class Root(Controller):
        @action
        def index(self, *a, **kw):
            yield ', '.join(a)

    with Given(Root()):
        assert status == 200
        assert response.text == ''

        when('/a')
        assert status == 200
        assert response.text == 'a'

        when('/a/b/c/d')
        assert status == 200
        assert response.text == 'a, b, c, d'


def test_fixed_positional_and_vary_optional_arguments_url_parameters():
    class Root(Controller):
        @action
        def index(self, a, b, **kw):
            yield f'{a}, {b}, {kw}'

    with Given(Root(), '/a/b'):
        assert status == 200
        assert response.text == 'a, b, {}'

        when('/a')
        assert status == 404

        when('/a/b/c')
        assert status == 404


def test_both_fixed_positional_and_optional_arguments():
    class Root(Controller):
        @action
        def index(self, a, b, c=None):
            yield f'{a}, {b}, {c}'

    with Given(Root(), '/a/b'):
        assert status == 200
        assert response.text == 'a, b, None'

        when('/a/b/c')
        assert status == 200
        assert response.text == 'a, b, c'

        when('/a')
        assert status == 404

        when('/a/b/c/d')
        assert status == 404


def test_fixed_positional_and_both_vary_and_fixed_optional_arguments():
    class Root(Controller):
        @action
        def index(self, a, b, c=None, **kw):
            yield f'{a}, {b}, {c}, {kw}'

    with Given(Root(), '/a/b'):
        assert status == 200
        assert response.text == 'a, b, None, {}'

        when('/a/b/c')
        assert status == 200
        assert response.text == 'a, b, c, {}'

        when('/a')
        assert status == 404

        when('/a/b/c/d')
        assert status == 404


def test_both_foxed_and_vary_positional_arguments():
    class Root(Controller):
        @action
        def index(self, a, b, *args):
            yield f'{a}, {b}, {", ".join(args)}'

    with Given(Root(), '/a/b'):
        assert status == 200
        assert response.text == 'a, b, '

        when('/a/b/c')
        assert status == 200
        assert response.text == 'a, b, c'

        when('/a')
        assert status == 404

        when('/a/b/c/d')
        assert status == 200






import time
import os
import re
import logging
from os.path import isdir, join, relpath, pardir, exists
from mimetypes import guess_type

from .exceptions import HTTPNotFound, HTTPMethodNotAllowed, HTTPForbidden, \
    HTTPStatus
from .contexts import context
from .constants import HTTP_DATETIME_FORMAT, UNLIMITED


logging.basicConfig(level=logging.INFO)



class Controller(object):
    """Base Controller
    """

    __nanohttp__ = dict(
        verbs=['any'],
        encoding='utf8',
        default_action='index',
        minimum_allowed_arguments=0,
        maximum_allowed_arguments=UNLIMITED
    )

    def _get_default_handler(self, remaining_paths):
        default_action = self.__nanohttp__['default_action']
        handler = getattr(self, default_action, None)
        if not handler:
            raise HTTPNotFound()

        return handler, remaining_paths

    def _find_handler(self, remaining_paths):
        if not remaining_paths or not hasattr(self, remaining_paths[0]):
            # Handler is not found, trying default handler
            return self._get_default_handler(remaining_paths)

        return getattr(self, remaining_paths[0], None), remaining_paths[1:]

    def _validate_handler(self, handler, remaining_paths):
        if not callable(handler) or not hasattr(handler, '__nanohttp__'):
            raise HTTPNotFound()

        manifest = handler.__nanohttp__
        min_arguments = manifest.get('minimum_allowed_arguments')
        max_arguments = manifest.get('maximum_allowed_arguments')
        args_len = len(remaining_paths)
        verbs = manifest.get('verbs', 'any')

        if min_arguments > args_len or \
                (max_arguments != UNLIMITED and max_arguments < args_len):
            raise HTTPNotFound()

        if verbs != ['any'] and context.method not in verbs:
            raise HTTPMethodNotAllowed()

        prevent_empty_form = manifest.get('prevent_empty_form')
        if prevent_empty_form and len(context.form) <= 0:
            raise HTTPStatus(
                prevent_empty_form \
                if isinstance(prevent_empty_form, str) \
                else '400 Empty Form'
            )

        prevent_form = manifest.get('prevent_form')
        if prevent_form and len(context.form) > 0:
            raise HTTPStatus(
                prevent_form \
                if isinstance(prevent_form, str) \
                else '400 Form Not Allowed'
            )

        form_whitelist = manifest.get('form_whitelist')
        if form_whitelist:
            if isinstance(form_whitelist, tuple):
                form_whitelist, status = form_whitelist
            else:
                status = None

            for k in context.form:
                if k not in form_whitelist:
                    raise HTTPStatus(
                        status or f'400 Field: {k} Not Allowed'
                    )

        return handler, remaining_paths

    def _serve_handler(self, handler, remaining_paths):
        context.response_encoding = handler.__nanohttp__.get('encoding', None)
        context.response_content_type = \
            handler.__nanohttp__.get('content_type', None)

        kwargs = {}
        for k, v in handler.__nanohttp__.get('keywordonly_arguments', []):
            value = context.query.get(k)
            if value:
                kwargs[k] = value

        return handler(*remaining_paths, **kwargs)

    def __call__(self, *remaining_paths):
        handler, remaining_paths = self._find_handler(list(remaining_paths))
        handler, remaining_paths = \
            self._validate_handler(handler, remaining_paths)
        return self._serve_handler(handler, remaining_paths)


class RestController(Controller):
    """HTTP method oriented controller
    """
    def _find_handler(self, remaining_paths):
        if remaining_paths and hasattr(self, remaining_paths[0]):
            return getattr(self, remaining_paths[0], None), remaining_paths[1:]

        # Handler is not found, trying verb
        if not hasattr(self, context.method):
            raise HTTPMethodNotAllowed()

        return getattr(self, context.method), remaining_paths


class Static(Controller):
    """Serves static files
    """
    __nanohttp__ = dict(
        verbs=['any'],
        encoding=None,
        default_action='index'
    )

    __chunk_size__ = 0x4000

    def __init__(self, directory='.', default_document='index.html'):
        """
        :param directory: Directory path to server
        :param default_document: Default document to serve as index
        """
        self.default_document = default_document
        self.directory = directory

    def __call__(self, *remaining_paths):

        # Find the physical path of the given path parts
        physical_path = join(self.directory, *remaining_paths)

        # Check to do not access the parent directory of root and also we are
        # not listing directories here.
        if pardir in relpath(physical_path, self.directory):
            raise HTTPForbidden()

        if isdir(physical_path):
            if not self.default_document:
                raise HTTPNotFound()

            physical_path = join(physical_path, self.default_document)
            if not (self.default_document and exists(physical_path)):
                raise HTTPNotFound()

        context.response_headers.add_header(
            'Content-Type',
            guess_type(physical_path)[0] or 'application/octet-stream'
        )

        try:
            f = open(physical_path, mode='rb')
            stat = os.fstat(f.fileno())
            context.response_headers.add_header('Content-Length', str(stat[6]))
            context.response_headers.add_header(
                'Last-Modified',
                time.strftime(HTTP_DATETIME_FORMAT, time.gmtime(stat.st_mtime))
            )

            with f:
                while True:
                    r = f.read(self.__chunk_size__)
                    if not r:
                        break
                    yield r

        except OSError:
            raise HTTPNotFound()


class RegexRouteController(Controller):
    """This is how to use it:

    .. code-block:: python

       class Root(RegexRouteController):

           def __init__(self):
               super().__init__((
                   (r'/installations/(?P<installation_id>\\d+)/access_tokens',
                   self.access_tokens),
               ))

           @json
           def access_tokens(self, installation_id: int):
               return dict(installationId=installation_id)

    """

    def __init__(self, routes):
        """
        :param routes: Routes list in (regex, method) format
        """
        self.routes = [(re.compile(p), a) for p, a in routes]

    def _find_handler(self, remaining_paths):
        path = '/' + '/'.join(remaining_paths)
        for pattern, handler in self.routes:
            match = pattern.match(path)
            if match:
                return handler, match.groups()
        raise HTTPNotFound()



from bddrest import status, response

from nanohttp import RegexRouteController, action
from nanohttp.tests.helpers import Given, when


def test_regex_controller():
    class Root(RegexRouteController):
        def __init__(self):
            super().__init__((
                (r'/foos/(?P<id>\d+)/bars',
                self.bars),
            ))

        @action
        def bars(self, id: int):
            return f'{id}'

    with Given(Root(), '/foos/1/bars'):
        assert status == 200
        assert response.text == '1'

        when('/foos')
        assert status == 404

        when('/foos/bars')
        assert status == 404

        when('/foos/1/2/bars')
        assert status == 404

        when('/foos/a/bars')
        assert status == 404




from nanohttp import LazyAttribute


def test_lazyattribute():
    global callcount
    class MyType:
        pass
    my_instance = MyType()

    callcount = 0
    class O:
        @LazyAttribute
        def attr1(self):
            """Attribute 1"""
            global callcount
            callcount += 1
            return my_instance

    assert 'Attribute 1' == O.attr1.__doc__
    assert 'attr1' == O.attr1.__name__

    o = O()
    assert my_instance is o.attr1
    assert my_instance is o.attr1
    assert my_instance is o.attr1
    assert 1 == callcount




from bddrest import status, response, given

from nanohttp import Application, Controller, action, html, json, text, \
    xml, binary, context
from nanohttp.tests.helpers import Given, when

app_init_calls = 0
begin_response_calls = 0
begin_request_calls = 0
end_response_calls = 0

def test_application_hooks():
    class Root(Controller):
        @action
        def index(self):
            foo = context.foo
            del context.foo
            context.bar = foo
            yield f'Index: {foo}'

    class TestApp(Application):
        def app_init(self):
            global app_init_calls
            app_init_calls += 1

        def begin_request(self):
            global begin_request_calls
            context.foo = 'Bar'
            begin_request_calls += 1

        def begin_response(self):
            global begin_response_calls
            assert not hasattr(context, 'foo')
            assert hasattr(context, 'bar')
            begin_response_calls += 1

        def end_response(self):
            global end_response_calls
            end_response_calls += 1

    with Given(TestApp(Root())):
        assert status == 200
        assert response.text == 'Index: Bar'
        assert app_init_calls == 1
        assert begin_response_calls == 1
        assert begin_request_calls == 1
        assert end_response_calls == 1

        when()
        assert status == 200
        assert app_init_calls == 1
        assert begin_response_calls == 2
        assert begin_request_calls == 2
        assert end_response_calls == 2




from bddrest import status, response

from nanohttp import Controller, action
from nanohttp.tests.helpers import Given, when


def test_object_dispatcher():
    class RegularClass:
        pass

    class BarController(Controller):
        @action
        def index(self, *args):
            yield f'Bars, {", ".join(args)}'

        def private(self):  # pragma: no cover
            raise Exception()

        @action
        def lorem(self, *args):
            yield f'Lorem, {", ".join(args)}'

    class Root(Controller):
        bars = BarController()
        requlars = RegularClass()

        @action
        def index(self):
            yield 'Index'

        @action
        def foo(self):
            yield 'Foo'

        def private(self):  # pragma: no cover
            raise Exception()

    with Given(Root()):
        assert status == 200
        assert response.text == 'Index'

        when('/foo')
        assert status == 200
        assert response.text == 'Foo'

        when(url='/a')
        assert status == '404 Not Found'

        when(url='/a/')
        assert status == '404 Not Found'

        when(url='/requlars')
        assert status == '404 Not Found'

        when('/private')
        assert status == 404

        when('/bars')
        assert status == 200
        assert response.text == 'Bars, '

        when('/bars/a')
        assert status == 200
        assert response.text == 'Bars, a'

        when('/bars/lorem')
        assert status == 200
        assert response.text == 'Lorem, '

        when('/bars/lorem/a/b')
        assert status == 200
        assert response.text == 'Lorem, a, b'

        when('/bars/private')
        assert status == 404




import pytest
from bddrest import status, response

from nanohttp import Controller, action, context
from nanohttp.tests.helpers import Given, when


def test_basic_pipeline():
    class Root(Controller):
        @action
        def index(self):
            yield f'Index: {context.request_scheme}, {context.request_uri}'

    with Given(Root()):
        assert status == 200
        assert response.text == 'Index: http, http://bddrest-interceptor/'


def test_no_default_handler():
    class Root(Controller):
        @action
        def foo(self):
            yield 'Foo'

    with Given(Root(), '/foo'):
        assert status == 200
        assert response.text == 'Foo'

        when('/')
        assert status == 404


def test_iterable_pipeline():
    class Root(Controller):
        @action
        def index(self):
            return ['a', 'b']

    with Given(Root()):
        assert status == 200
        assert response.text == 'ab'


def test_not_iterable_response():
    class Root(Controller):
        @action
        def index(self):
            return 352345352

    with pytest.raises(ValueError):
        Given(Root())


def test_iterable_but_bad_response():
    class Root(Controller):
        @action
        def index(self):
            yield '1'
            yield 352345352

    with pytest.raises(TypeError):
        Given(Root())


def test_empty_response():
    class Root(Controller):
        @action
        def index(self):
            return None

    with Given(Root()):
        assert status == 200
        assert response.text == ''




from http import cookies


import pytest
from bddrest import status, response

from nanohttp import Controller, action, context
from nanohttp.tests.helpers import Given, when


def test_basic_pipeline():
    class Root(Controller):
        @action
        def index(self):
            counter = context.cookies['counter']
            context.cookies['counter'] = str(int(counter.value) + 1)
            context.cookies['counter']['max-age'] = 1
            context.cookies['counter']['path'] = '/a'
            context.cookies['counter']['domain'] = 'example.com'
            yield 'Index'

    headers = {'Cookie': 'counter=1;'}
    with Given(Root(), headers=headers):
        assert status == 200
        assert response.text == 'Index'
        assert 'Set-cookie' in response.headers
        cookie = cookies.SimpleCookie(response.headers['Set-cookie'])
        counter = cookie['counter']
        assert counter.value == '2'
        assert counter['path'] == '/a'
        assert counter['domain'] == 'example.com'
        assert counter['max-age'] == '1'




from bddrest import status, response

from nanohttp import RestController, action
from nanohttp.tests.helpers import Given, when


def test_rest_controller():
    class Root(RestController):

        @action
        def index(self):  # pragma: no cover
            raise Exception()

        @action
        def foo(self, *args):
            yield f'Foo, {", ".join(args)}'

        def private(self):  # pragma: no cover
            raise Exception()

    with Given(Root(), verb='foo'):
        assert status == 200
        assert response.text == 'Foo, '

        when(verb='FOO')
        assert status == 200
        assert response.text == 'Foo, '

        when('/a', verb='FOO')
        assert status == 200
        assert response.text == 'Foo, a'

        when('/a/b', verb='FOO')
        assert status == 200
        assert response.text == 'Foo, a, b'

        when(verb='GET')
        assert status == 405

        when('/private')
        assert status == 404

        when('/private', verb='GET')
        assert status == 404


def test_nested_rest_controllers():
    class BarController(RestController):
        @action
        def get(self, *args):
            yield f'Bars, {", ".join(args)}'

    class Root(RestController):
        bars = BarController()

        @action
        def get(self, *args):
            yield f'{", ".join(args)}'

    with Given(Root()):
        assert status == 200
        assert response.text == ''

        when('/bars')
        assert status == 200
        assert response.text == 'Bars, '

        when('/bars/a')
        assert status == 200
        assert response.text == 'Bars, a'





from pymlconf import DeferredRoot


settings = DeferredRoot()


BUILTIN_CONFIGURATION = """
debug: true
domain:
cookie:
  http_only: false
  secure: false
"""


def configure(*args, **kwargs):
    """Load configurations

    .. seealso:: `pymlconf Documentations
                 <https://github.com/pylover/pymlconf#documentation>`_

    :param args: positional arguments pass into
                 ``pymlconf.DeferredRoot.load``
    :param kwargs: keyword arguments pass into
                   ``pymlconf.DeferredRoot.load``
    """

    settings.initialize(BUILTIN_CONFIGURATION, **kwargs)
    for a in args:
        settings.merge(a)





import io
import cgi

from bddrest import status, response

from nanohttp import Controller, action, context
from nanohttp.tests.helpers import Given, when


def test_http_url_encoded_form():
    class Root(Controller):
        @action
        def index(self):
            yield context.request_content_type
            yield ', '
            yield ', '.join(
                f'{k}={v}' for k, v in sorted(context.form.items())
            )

    with Given(Root(), verb='POST'):
        assert status == 200
        assert response.text == ', '

        when(form=dict(a=1, b=2))
        assert status == 200
        assert response.text == 'application/x-www-form-urlencoded, a=1, b=2'

        when(body='a')
        assert status == 200
        assert response.text == ', a='

        # Duplicated fields
        when(body='a=1&a=2', content_type='application/x-www-form-urlencoded')
        assert status == 200
        assert response.text == \
            'application/x-www-form-urlencoded, a=[\'1\', \'2\']'


def test_http_multipart_form():
    class Root(Controller):
        @action
        def index(self):
            def read_form_field(v):
                return v.file.read().decode() \
                    if isinstance(v, cgi.FieldStorage) else v
            yield context.request_content_type
            yield ', '
            yield ', '.join(
                f'{k}='
                f'{read_form_field(v)}' \
                for k, v in sorted(context.form.items())
            )

    with Given(Root(), verb='POST'):
        assert status == 200
        assert response.text == ', '

        when(multipart=dict(a=1, b=2))
        assert status == 200
        assert response.text == 'multipart/form-data, a=1, b=2'

        filelike = io.BytesIO(b'abcdef')
        when(multipart=dict(a=filelike))
        assert status == 200
        assert response.text == 'multipart/form-data, a=abcdef'


def test_json_form():
    class Root(Controller):
        @action
        def index(self):
            form = context.form
            yield context.request_content_type
            yield ', '
            yield ', '.join(
                f'{k}={v}' for k, v in sorted(form.items())
            )

    with Given(Root(), verb='POST', json={}, configuration='debug: false'):
        assert status == 200
        assert response.text == ', '

        when(json=dict(a=1, b=2))
        assert status == 200
        assert response.text == 'application/json, a=1, b=2'

        # No content length
        when(body='{}', content_type='application/json')
        assert status == 400
        assert response.text == 'Content-Length required'

        when(
            body='malformed',
            content_type='application/json',
            headers={'Content-Length': '9'}
        )
        assert status == 400
        assert response.text == 'Cannot parse the request'


def test_invalid_form():
    class Root(Controller):
        @action
        def index(self):
            form = context.form
            yield context.request_content_type
            yield ', '
            yield ', '.join(
                f'{k}={v}' for k, v in sorted(form.items())
            )

    with Given(Root(), verb='POST', json={}):
        assert status == 200
        assert response.text == ', '

        when(body='', content_type='multipart/form-data; boundary=')
        assert status == '400 Cannot parse the request'




import ujson
import traceback

from .contexts import context
from .configuration import settings


class HTTPStatus(Exception):
    status = None
    headers = None

    def __init__(self, status=None):
        if status is not None:
            self.status = status

        super().__init__(self.status)

    def render(self):
        stack_trace = traceback.format_exc()
        if context.response_content_type == 'application/json':
            return ujson.encode(
                dict(stackTrace=stack_trace) if settings.debug else dict()
            )
        context.response_encoding = 'utf-8'
        context.response_content_type = 'text/plain'
        return stack_trace if settings.debug \
            else self.status.split(' ', 1)[1]

    @property
    def headers(self):
        if context.response_content_type == 'application/json':
            contenttype = context.response_content_type
        else:
            contenttype = 'text/plain;charset=utf-8'

        return [('ContentType', contenttype)]


class HTTPKnownStatus(HTTPStatus):
    def __init__(self, status_text=None):
        code, text = self.status.split(' ', 1)
        super().__init__(f'{code} {status_text or text}')


class HTTPBadRequest(HTTPKnownStatus):
    status = '400 Bad Request'


class HTTPUnauthorized(HTTPKnownStatus):
    status = '401 Unauthorized'


class HTTPForbidden(HTTPKnownStatus):
    status = '403 Forbidden'


class HTTPNotFound(HTTPKnownStatus):
    status = '404 Not Found'


class HTTPMethodNotAllowed(HTTPKnownStatus):
    status = '405 Method Not Allowed'


class HTTPConflict(HTTPKnownStatus):
    status = '409 Conflict'


class HTTPGone(HTTPKnownStatus):
    status = '410 Gone'


class HTTPPreconditionFailed(HTTPKnownStatus):
    status = '412 Precondition Failed'


class HTTPRedirect(HTTPKnownStatus):
    """
    This is an abstract class for all redirects.
    """

    def __init__(self, location, *args, **kw):
        self._headers = [('Location', location)]
        super().__init__(*args, **kw)

    @property
    def headers(self):
        headers = super().headers
        return headers + self._headers


class HTTPMovedPermanently(HTTPRedirect):
    status = '301 Moved Permanently'


class HTTPFound(HTTPRedirect):
    status = '302 Found'


class HTTPNotModified(HTTPKnownStatus):
    status = '304 Not Modified'


class HTTPInternalServerError(HTTPKnownStatus):
    status = '500 Internal Server Error'


class HTTPBadGatewayError(HTTPKnownStatus):
    status = '502 Bad Gateway'


class HTTPSuccess(HTTPKnownStatus):
    status = '200 OK'


class HTTPCreated(HTTPSuccess):
    status = '201 Created'


class HTTPAccepted(HTTPSuccess):
    status = '202 Accepted'


class HTTPNonAuthoritativeInformation(HTTPSuccess):
    status = '203 Non-Authoritative Information'


class HTTPNoContent(HTTPSuccess):
    status = '204 No Content'


class HTTPResetContent(HTTPSuccess):
    status = '205 Reset Content'


class HTTPPartialContent(HTTPSuccess):
    status = '206 Partial Content'



import re
import functools
from _decimal import InvalidOperation

from nanohttp import context
from nanohttp.exceptions import HTTPStatus, HTTPBadRequest


NO_EMPTY_FORM = 'NO_EMPTY_FORM'
NO_FORM = 'NO_FORM'


class Field:
    def __init__(self, title, form=True, query=False, required=None,
                 type_=None, minimum=None, maximum=None, pattern=None,
                 min_length=None, max_length=None, callback=None,
                 not_none=None, readonly=None):
        self.title = title
        self.form = form
        self.query = query
        self.criteria = []

        if readonly:
            self.criteria.append(ReadonlyValidator(readonly))

        if required:
            self.criteria.append(RequiredValidator(required))

        if not_none:
            self.criteria.append(NotNoneValidator(not_none))

        if type_:
            self.criteria.append(TypeValidator(type_))

        if minimum:
            self.criteria.append(MinimumValidator(minimum))

        if maximum:
            self.criteria.append(MaximumValidator(maximum))

        if pattern:
            self.criteria.append(PatternValidator(pattern))

        if min_length:
            self.criteria.append(MinLengthValidator(min_length))

        if max_length:
            self.criteria.append(MaxLengthValidator(max_length))

        if callback:
            self.criteria.append(CallableValidator(callback))

    def validate(self, container):
        for criterion in self.criteria:
            criterion.validate(self, container)

        return container


class Criterion:
    def __init__(self, expression):
        if isinstance(expression, tuple):
            error = expression[1]
            self.expression = expression[0]
        else:
            self.expression = expression
            error = '400 Bad request'

        parsed_error = error.split(' ', 1)

        self.status_code = int(parsed_error[0])

        if len(parsed_error) == 2:
            self.status_text = parsed_error[1]
        else:
            self.status_text = 'Bad request'

    def validate(self, field: Field, container: dict) -> None:
        value = container.get(field.title)
        if value is None:
            return

        container[field.title] = self._validate(
            container[field.title],
            container,
            field
        )

    def _validate(self, value, container: dict, field: Field
                  ):  # pragma: no cover
        """
        It must be overridden in the child class.

        This method should raise exception if the criterion is not met. there
        is a chance to set
        a new value because the
        container is available here.
        :param value: The value to validate
        :param field:
        :param container:
        :return:
        """
        raise NotImplementedError()

    def create_exception(self):
        if self.status_code == 400:
            return HTTPBadRequest(self.status_text)

        return HTTPStatus(
            status=f'{self.status_code} {self.status_text}'
        )


class FlagCriterion(Criterion):
    def __init__(self, expression):
        if isinstance(expression, bool):
            error = '400 Bad request'
        elif isinstance(expression, str):
            error = expression
        else:
            raise TypeError('Only bool and or string will be accepted.')

        super().__init__((True, error))


class RequiredValidator(FlagCriterion):
    def validate(self, field, container):
        if field.title not in container:
            raise self.create_exception()


class NotNoneValidator(FlagCriterion):
    def validate(self, field, container):
        if field.title not in container:
            return

        if container[field.title] is None:
            raise self.create_exception()


class ReadonlyValidator(FlagCriterion):
    def validate(self, field, container):
        if field.title in container:
            raise self.create_exception()


class TypeValidator(Criterion):

    def _validate(self, value, container, field):
        type_ = self.expression
        try:
            return type_(value)
        except (ValueError, TypeError, InvalidOperation):
            raise self.create_exception()


class MinLengthValidator(Criterion):

    def _validate(self, value, container, field):
        if len(value) < self.expression:
            raise self.create_exception()

        return value


class MaxLengthValidator(Criterion):

    def _validate(self, value, container, field):
        if len(value) > self.expression:
            raise self.create_exception()

        return value


class MinimumValidator(Criterion):

    def _validate(self, value, container, field):
        try:
            if value < self.expression:
                raise self.create_exception()
        except TypeError:
            raise self.create_exception()

        return value


class MaximumValidator(Criterion):

    def _validate(self, value, container, field):
        try:
            if value > self.expression:
                raise self.create_exception()
        except TypeError:
            raise self.create_exception()

        return value


class PatternValidator(Criterion):

    def _validate(self, value, container, field):
        pattern = re.compile(self.expression) \
            if isinstance(self.expression, str) \
            else self.expression
        if pattern.match(value) is None:
            raise self.create_exception()

        return value


class RequestValidator:
    def __init__(self, fields, empty_form=None):
        # Merging default specification
        self.fields = {}
        self.empty_form = empty_form
        for field_name, specification in fields.items():
            kwargs = dict(callback=specification) \
                if callable(specification) else specification

            self.fields[field_name] = Field(field_name, **kwargs)

    def __call__(self, form=None, query=None, *args, **kwargs):
        for field_name, field in self.fields.items():

            if query and field.query:
                field.validate(query)

            if field.form \
                    and field.title is not None \
                    and (query is None or field.title not in query):
                field.validate(form)

        return form, query

class CallableValidator(Criterion):

    def _validate(self, value, container, field):
        return self.expression(value, container, field)


def validate(**fields):
    """Decorator to validate HTTP Forms and query string.

    .. code-block:: python

       @validate(field1=dict(required=True))
       def index(self, *, field1):
           ...


    Available parameters for validation is listed below:

    :param required: Boolean or str, indicates the field is required.
    :param not_none: Boolean or str, Raise when field is given and it's value
                     is None.

    :param type_: A callable to pass the received value to it as the only
                  argument and get it in the apprpriate type, Both
                  ``ValueError`` and `TypeError`` may be raised if the value
                  cannot casted to the specified type. A good example of this
                  callable would be the :class:`int`.
    :param minimum: Numeric, Minimum allowed value.
    :param maximum: Numeric, Maximum allowed value.
    :param pattern: Regex pattern to match the value.
    :param min_length: Only for strings, the minumum allowed length of the
                       value.
    :param max_length: Only for strings, the maximum allowed length of the
                       value.
    :param callback: A ``callable(value, container, field: Field)`` to be
                     called while validating the field.


    A detailed example:

    .. code-block:: python

       my_validator = validate(
           title=dict(
               required='710 Title not in form',
               max_length=(50, '704 At most 50 characters are valid for title')
           ),
           description=dict(
               max_length=(
                   512,
                   '703 At most 512 characters are valid for description'
               )
           ),
           dueDate=dict(
               pattern=(DATE_PATTERN, '701 Invalid due date format'),
               required='711 Due date not in form'
           ),
           cutoff=dict(
               pattern=(DATE_PATTERN, '702 Invalid cutoff format'),
               required='712 Cutoff not in form'
           ),
       )

       @json
       @my_validator
       def index(self):
           ...


    """

    validator = RequestValidator(fields)

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            validator(form=context.form, query=context.query)
            return func(*args, **kwargs)

        return wrapper

    return decorator




import pytest
from bddrest import status, response

from nanohttp import Controller, action, HTTPStatus, json, HTTPBadRequest, \
    HTTPMovedPermanently, HTTPNotModified
from nanohttp.tests.helpers import Given


def test_http_status_debug_mode():
    class Root(Controller):
        @action
        def index(self):
            raise HTTPStatus('603 Bad Happened')

    with Given(Root()):
        assert status == '603 Bad Happened'
        assert 'traceback' in response.text.casefold()


def test_http_status_no_debug_mode():
    class Root(Controller):
        @action
        def index(self):
            raise HTTPStatus('603 Bad Happened')

    with Given(Root(), configuration='debug: false'):
        assert status == '603 Bad Happened'
        assert 'traceback' not in response.text.casefold()
        assert response.text == 'Bad Happened'


def test_http_status_debug_mode_json_content_type():
    class Root(Controller):
        @json
        def index(self):
            raise HTTPStatus('603 Bad Happened')

    with Given(Root()):
        assert status == '603 Bad Happened'
        assert 'stackTrace' in response.json


def test_http_status_no_debug_mode_json_content_type():
    class Root(Controller):
        @json
        def index(self):
            raise HTTPStatus('603 Bad Happened')

    with Given(Root(), configuration='debug: false'):
        assert status == '603 Bad Happened'
        assert response.json == {}


def test_http_bad_request():
    class Root(Controller):
        @action
        def index(self):
            raise HTTPBadRequest('My Bad')

    with Given(Root(), configuration='debug: false'):
        assert status == '400 My Bad'
        assert response.text == 'My Bad'


def test_http_not_modified():
    class Root(Controller):
        @action
        def index(self):
            raise HTTPNotModified()

    with Given(Root()):
        assert status == 304


def test_http_redirect():
    class Root(Controller):
        @action
        def index(self):
            raise HTTPMovedPermanently('http://example.com')

    with Given(Root(), configuration='debug: false'):
        assert status == '301 Moved Permanently'
        assert response.text == 'Moved Permanently'
        assert response.headers['Location'] == 'http://example.com'


def test_unhandled_exceptions_debug_mode():
    class E(Exception):
        pass

    class Root(Controller):
        @action
        def index(self):
            raise E()

    with pytest.raises(E):
        Given(Root())


def test_unhandled_exceptions_no_debug_mode():
    class E(Exception):
        pass

    class Root(Controller):
        @action
        def index(self):
            raise E()

    with pytest.raises(E):
        Given(Root(), configuration='debug: false')





import requests

from nanohttp import Application, Controller, action


def test_quickstart_without_arguments(run_quickstart):
    url = run_quickstart()
    response = requests.get(url)
    assert response.status_code == 200


def test_quickstart_with_initial_config(run_quickstart):
    url = run_quickstart(config='debug: false')
    response = requests.get(url)
    assert response.status_code == 200


def test_quickstart_with_application(run_quickstart):
    class Root(Controller):
        @action
        def index(self):
            yield f'Index'

    url = run_quickstart(application=Application(Root()))
    response = requests.get(url)
    assert response.status_code == 200
    assert response.text == 'Index'




from bddrest import status, response

from nanohttp import Controller, action, context
from nanohttp.tests.helpers import Given, when


def test_query_strign():
    class Root(Controller):
        @action
        def index(self, *, a=1, b=2):
            query = ', '.join(f'{k}={v}' for k, v in context.query.items())
            yield f'{a}, {b}, {query}'

    with Given(Root()):
        assert status == 200
        assert response.text == '1, 2, '

        when(query=dict(a='a'))
        assert status == 200
        assert response.text == 'a, 2, a=a'

        when(query=dict(a='a', c='c'))
        assert status == 200
        assert response.text == 'a, 2, a=a, c=c'

        when(query=dict(c='c'))
        assert status == 200
        assert response.text == '1, 2, c=c'




import threading
import wsgiref.util
import wsgiref.headers

from typing import Union
from urllib.parse import parse_qs
from http.cookies import SimpleCookie

from nanohttp import exceptions
from .helpers import LazyAttribute, parse_any_form


class ContextIsNotInitializedError(Exception):
    pass


class ContextStack(list):

    def push(self, item):
        self.append(item)


class Context:
    """A Global context for Request and Response.

    Context are initialized and entering on every request (referring to
     nanohttp application lifecycle).

    Context also supports to use stack nested (>=0.16.6), its useful on
    testing.

    .. code-block:: python

        def sample:
            return context.query['weather']

        with Context({'QUERY_STRING': 'weather=Sunny'}):
            assert sample() == 'Sunny

            with Context({'QUERY_STRING': 'weather=Rainy'}):
                assert sample() == 'Rainy'

    """

    #: Response encoding
    response_encoding = None

    #: Thread local variable contexts stored in
    thread_local = threading.local()

    #: Current :class:`.Application` instance
    application = None

    @property
    def __stack__(self):
        """Nested contexts stack
        """
        THREADLOCAL_STACK_ATTRIBUTE = 'nanohttp_context_stack'
        if not hasattr(self.thread_local, THREADLOCAL_STACK_ATTRIBUTE):
            setattr(
                self.thread_local, THREADLOCAL_STACK_ATTRIBUTE, ContextStack()
            )
        return getattr(self.thread_local, THREADLOCAL_STACK_ATTRIBUTE)

    def __init__(self, environ, application=None):
        """
        :param environ: WSGI environ dictionary
        :param application: :class:`.Application` instance
        """
        self.environ = environ
        self.application = application
        self.response_headers = wsgiref.headers.Headers()

    def __enter__(self):
        # Backing up the current context
        if hasattr(self.thread_local, 'nanohttp_context'):
            self.__stack__.push(self.thread_local.nanohttp_context)

        self.thread_local.nanohttp_context = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.thread_local.nanohttp_context
        if self.__stack__:
            self.thread_local.nanohttp_context = self.__stack__.pop()

    @LazyAttribute
    def request_content_length(self) -> Union[int, None]:
        """Request content length
        """
        v = self.environ.get('CONTENT_LENGTH')
        return None if not v or not v.strip() else int(v)

    @LazyAttribute
    def request_content_type(self) -> Union[str, None]:
        """Request content type
        """
        content_type = self.environ.get('CONTENT_TYPE')
        if content_type:
            return content_type.split(';')[0]
        return None

    @property
    def response_content_type(self) -> Union[str, None]:
        """Response content type property
        """
        content_type = self.response_headers.get('Content-Type')
        if content_type:
            return content_type.split(';')[0]
        return None

    @response_content_type.setter
    def response_content_type(self, v):
        if v is None:
            del self.response_headers['Content-Type']
        else:
            self.response_headers['Content-Type'] = \
                '%s; charset=%s' % (v, self.response_encoding)

    @classmethod
    def get_current(cls) -> 'Context':
        """Get current context

            Not initialized context raises
            :class:`.ContextIsNotInitializedError`,
        """
        if not hasattr(cls.thread_local, 'nanohttp_context'):
            raise ContextIsNotInitializedError(
                "Context is not initialized yet."
            )
        return cls.thread_local.nanohttp_context

    @LazyAttribute
    def method(self):
        """`HTTP Request method
        <https://tools.ietf.org/html/rfc7231#section-4.3>`_
        """
        return self.environ['REQUEST_METHOD'].lower()

    @LazyAttribute
    def path(self):
        """Request path
        """
        return self.environ['PATH_INFO']

    @LazyAttribute
    def request_uri(self):
        """Request full URI (includes query string)
        """
        return wsgiref.util.request_uri(self.environ, include_query=True)

    @LazyAttribute
    def request_scheme(self):
        """Request Scheme (http|https)
        """
        return wsgiref.util.guess_scheme(self.environ)

    @LazyAttribute
    def query(self):
        """Request query string
        """
        if 'QUERY_STRING' not in self.environ:
            return {}

        return {k: v[0] if len(v) == 1 else v for k, v in parse_qs(
            self.environ['QUERY_STRING'],
            keep_blank_values=True,
            strict_parsing=False
        ).items()}

    @LazyAttribute
    def form(self):
        """Request form values

        .. note:: if using `multipart/form-data` uploaded file will reproduce
         as ``cgi.FieldStorage``.
        """
        return parse_any_form(
            self.environ,
            content_length=self.request_content_length,
            content_type=self.request_content_type
        )

    @LazyAttribute
    def cookies(self) -> 'SimpleCookie':
        """Cookies
        """
        result = SimpleCookie()
        if 'HTTP_COOKIE' in self.environ:
            result.load(self.environ['HTTP_COOKIE'])
        return result

    def encode_response(self, buff):
        """Encode response buffer with encoding definition on current
        context
        """
        try:
            if self.response_encoding:
                return buff.encode(self.response_encoding)
            else:
                return buff
        except AttributeError:
            raise TypeError(
                'The returned response should has the `encode` attribute, '
                'such as `str`.'
            )


class ContextProxy(Context):

    # noinspection PyInitNewSignature
    def __new__(cls) -> Context:
        type_proxy = type('ContextProxy', (object, ), {
            '__getattr__': cls.__getattr__,
            '__setattr__': cls.__setattr__,
            '__delattr__': cls.__delattr__
        })
        # noinspection PyTypeChecker
        return object.__new__(type_proxy)

    def __getattr__(self, key):
        return getattr(Context.get_current(), key)

    def __setattr__(self, key, value):
        setattr(Context.get_current(), key, value)

    def __delattr__(self, key):
        delattr(Context.get_current(), key)


context = ContextProxy()



from bddrest import status, response

from nanohttp import Controller, action, context, chunked
from nanohttp.tests.helpers import Given, when


def test_chunked_streaming():
    class Root(Controller):
        @action
        @chunked
        def index(self):
            yield 'first'
            yield 'second'

        @action
        @chunked('trailer1', 'end')
        def trailer(self):
            yield 'first'
            yield 'second'

        @action
        @chunked
        def error(self):
            yield 'first'
            raise Exception('error in streaming')

    with Given(Root()):
        assert status == 200
        assert response.text == \
            '5\r\nfirst\r\n6\r\nsecond\r\n0\r\n\r\n'

        when('/trailer')
        assert status == 200
        assert response.text == \
            '5\r\nfirst\r\n6\r\nsecond\r\n0\r\ntrailer1: end\r\n\r\n'

        when('/error')
        assert status == 200
        assert response.text == \
            '5\r\nfirst\r\n18\r\nerror in streaming0\r\n\r\n'




import sys
import signal
from os import chdir
from os.path import relpath, basename

import nanohttp
import yaml
from .helpers import quickstart, load_controller_from_file
from .configuration import configure, settings

DEFAULT_CONFIG_FILE = 'nanohttp.yml'
DEFAULT_ADDRESS = '8080'


def parse_arguments(argv=None):
    import argparse

    argv = argv or sys.argv

    parser = argparse.ArgumentParser(prog=basename(argv[0]))
    parser.add_argument(
        '-c',
        '--config-file',
        action='append',
        default=[],
        dest='config_files',
        help='This option may be passed multiple times.'
    )

    parser.add_argument(
        '-b',
        '--bind',
        default=DEFAULT_ADDRESS,
        metavar='{HOST:}PORT',
        help='Bind Address. default: %s' % DEFAULT_ADDRESS
    )

    parser.add_argument(
        '-C',
        '--directory',
        default='.',
        help='Change to this path before starting the server default is: `.`'
    )

    parser.add_argument(
        '-V',
        '--version',
        default=False,
        action='store_true',
        help='Show the version.'
    )

    parser.add_argument(
        'controller',
        nargs='?',
        metavar='{MODULE{.py}}{:CLASS}',
        help='The python module and controller class to launch. default is '
             'python built-in\'s : `demo_app`, And the default value for '
             '`:CLASS` is `:Root` if omitted.'
    )

    parser.add_argument(
        '-o', '--option',
        action='append',
        metavar='key1.key2=value',
        dest='options',
        default=[],
        help= \
            'Configuration value to override. this option could be passed ' \
            'multiple times.'
    )

    return parser.parse_args(argv[1:])


def main(argv=None):
    args = parse_arguments(argv=argv)

    if args.version:  # pragma: no cover
        print(nanohttp.__version__)
        return 0

    def signal_handler(signum, frame):
        print('SIGTERM received')
        sys.exit(1)

    signal.signal(signal.SIGTERM, signal_handler)

    try:
        host, port = args.bind.split(':')\
            if ':' in args.bind else ('', args.bind)

        # Change dir
        if relpath(args.directory, '.') != '.':
            chdir(args.directory)

        configure(force=True)

        for f in args.config_files:
            settings.load_file(f)


        try:
            for option in args.options:
                key, value = option.split('=')
                value = yaml.load(value)
                if isinstance(value, str):
                    value = f'"{value}"'

                exec(f'settings.{key} = {value}')
        except AttributeError:
            print(f'Invalid configuration option: {key}', file=sys.stderr)
            return 1

        quickstart(
            controller=load_controller_from_file(args.controller),
            host=host,
            port=int(port)
        )
    except KeyboardInterrupt:  # pragma: no cover
        print('CTRL+C detected.')
        return 1
    else:  # pragma: no cover
        return 0





from setuptools import setup, find_packages
import os.path
import re

# reading package's version (same way sqlalchemy does)
with open(
    os.path.join(os.path.dirname(__file__), 'nanohttp', '__init__.py')
) as v_file:
    package_version = \
        re.compile('.*__version__ = \'(.*?)\'', re.S)\
        .match(v_file.read())\
        .group(1)


dependencies = [
    'pymlconf >= 1.0.1',
    'ujson'
]


setup(
    name='nanohttp',
    version=package_version,
    author='Vahid Mardani',
    author_email='vahid.mardani@gmail.com',
    url='http://github.com/Carrene/nanohttp',
    description='A very micro http framework.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',  # This is important!
    install_requires=dependencies,
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'nanohttp = nanohttp:main'
        ]
    },
    license='MIT',
    classifiers=[
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Development Status :: 4 - Beta',
        'License :: Other/Proprietary License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ]
    )



from os import path

import requests


def test_cli_without_arguments(clitool):
    url = clitool.execute()
    response = requests.get(url)
    assert response.status_code == 200
    assert len(response.text) > 1


# Erorred because of the travis buf about change ditectory:
# https://github.com/pytest-dev/pytest/issues/2686
def test_cli_with_change_directory_flag(clitool):
    url = clitool.execute('-C', '/tmp')
    response = requests.get(url)
    assert response.status_code == 200
    assert len(response.text) > 1


def test_cli_with_given_module_name(clitool, controller_file):
    directory, filename = path.split(controller_file())
    # Removing the module[.py] extension by filename[:-3]
    url = clitool.execute('-C', directory, filename[:-3])
    response = requests.get(url)
    assert response.status_code == 200
    assert response.text == 'Index'


def test_cli_with_change_dir_and_controller_filename(clitool, controller_file):
    directory, filename = path.split(controller_file())
    url = clitool.execute('-C', directory, filename)
    response = requests.get(url)
    assert response.status_code == 200
    assert response.text == 'Index'


def test_cli_with_static_controller(clitool, make_temp_directory):
    directory = make_temp_directory(a='Hello A')
    url = clitool.execute('-C', directory, ':Static')
    response = requests.get(f'{url}/a')
    assert response.status_code == 200
    assert response.text == 'Hello A'


def test_cli_with_config_file_flag(clitool, make_temp_file):
    filename = make_temp_file()
    with open(filename, 'w') as f:
        f.write('debug: false')

    url = clitool.execute('-c', filename)
    response = requests.get(url)
    assert response.status_code == 200
    assert len(response.text) > 1


def test_cli_with_option_flag(clitool):
    url = clitool.execute('-o', 'debug=false', '-o', 'foo=bar')
    response = requests.get(url)
    assert response.status_code == 200
    assert len(response.text) > 1


def test_cli_with_invalid_option_flag(clitool):
    url = clitool.execute('-o', 'a.b=1' )
    assert clitool.exitstatus != 0


def test_cli_with_given_filename(clitool, controller_file):
    url = clitool.execute(controller_file())
    response = requests.get(url)
    assert response.status_code == 200
    assert response.text == 'Index'


def test_cli_with_given_filename_and_controller_name(clitool, controller_file):
    url = clitool.execute(f'{controller_file()}:Root')
    response = requests.get(url)
    assert response.status_code == 200
    assert response.text == 'Index'

def test_cli_with_given_package_and_controller_name(clitool, controller_file):
    directory, filename = path.split(controller_file('__init__.py'))
    url = clitool.execute(f'{directory}:Root')
    response = requests.get(url)
    assert response.status_code == 200
    assert response.text == 'Index'




import os
import sys
import time
import shutil
import socket
import signal
import tempfile
from os import path
from multiprocessing import Process

import pytest
from pytest_cov.embed import cleanup

from nanohttp import main, quickstart, settings


@pytest.fixture
def make_temp_file():
    temp_files = []

    def _make_temp_file(**kw):
        """
        Structure example: {'a.html': 'Hello', 'b': {}}
        """
        filename = tempfile.mktemp()
        temp_files.append(filename)
        return filename

    yield _make_temp_file

    for f in temp_files:
        os.remove(f)


@pytest.fixture
def make_temp_directory():
    temp_directories = []

    def create_nodes(root, **kw):
        for k, v in kw.items():
            name = path.join(root, k)

            if isinstance(v, dict):
                os.mkdir(name)
                create_nodes(name, **v)
                continue

            with open(name, 'w') as f:
                f.write(v)


    def _make_temp_directory(**kw):
        """
        Structure example: {'a.html': 'Hello', 'b': {}}
        """
        root = tempfile.mkdtemp()
        temp_directories.append(root)
        create_nodes(root, **kw)
        return root

    yield _make_temp_directory

    for d in temp_directories:
        shutil.rmtree(d)


@pytest.fixture
def free_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.bind((socket.gethostname(), 0))
        return s.getsockname()[1]
    finally:
        s.close()


@pytest.fixture
def clitool(free_port):
    class Tool:
        subprocess = None

        def terminate_subprocess(self, sig, frame):  # pragma: no cover
            cleanup()
            sys.exit(sig)

        def wrapper(self, args):
            signal.signal(signal.SIGTERM, self.terminate_subprocess)
            sys.exit(main(args))

        def execute(self, *a):
            port = free_port
            args = ['nanohttp', f'-b{port}']
            args.extend(a)
            self.subprocess = Process(
                target=self.wrapper,
                args=(args, ),
                daemon=True
            )
            self.subprocess.start()
            time.sleep(.5)
            return f'http://localhost:{port}'

        def _wait_for(self):
            if self.subprocess.is_alive():
                self.subprocess.terminate()
            self.subprocess.join()

        def terminate(self):
            if self.subprocess is not None:
                self._wait_for()
                self.subprocess = None

        @property
        def exitstatus(self):
            self._wait_for()
            return self.subprocess.exitcode

    tool = Tool()
    yield tool
    tool.terminate()


@pytest.fixture
def controller_file(make_temp_directory):
    def _create(name='foo.py'):
        directory = make_temp_directory()
        filename = path.join(directory, name)
        with open(filename, 'w') as f:
            f.write(
                'from nanohttp import Controller, action\n'
                'class Root(Controller):\n'
                '    @action\n'
                '    def index(self):\n'
                '        yield \'Index\'\n'
            )
        return filename
    yield _create


@pytest.fixture
def run_quickstart(free_port):
    terminators = []
    settings.__class__._instance = None
    def wrapper(*args, **kw):
        terminate = quickstart(*args, block=False, port=free_port, **kw)
        time.sleep(.1)
        terminators.append(terminate)
        return f'http://localhost:{free_port}'
    yield wrapper
    for t in terminators:
        t()
        time.sleep(.1)


